unit IndexList;

interface

uses
  {$IFDEF MINIMAL}MiniClasses{$ELSE}Classes{$ENDIF}, SysUtils, Windows;

type
  TIndexRec = record
    Index: Integer;
    Data: Pointer;
  end;
  PIndexRec = ^TIndexRec;

type
  TEachItemProc = procedure(Sender: TObject; const Data: Pointer; const Index: Integer) of object;

type
  EIndexListError = class(Exception);

type
  TIndexList = class(TList)
  private
    FListLock: PRTLCriticalSection;
    function    FindNativeIndex(const Index: Integer): Integer;
  protected
    function    Get_FirstIndex: Integer;
    function    Get_LastIndex: Integer;
    function    Get_NextIndex(const CurIndex: Integer): Integer;
    function    Get_PrevIndex(const CurIndex: Integer): Integer;

    function    GetLocked: boolean;
    procedure   SetLocked(const Value: boolean);
    procedure   EnterCriticalSection;
    procedure   LeaveCriticalSection;
  public
    constructor Create;
    destructor  Destroy; override;
    function    Add(const Data: Pointer): Integer;
    procedure   Insert(const Index: Integer; const Data: Pointer);
    procedure   Delete(const Index: Integer);
    function    FirstIndex: Integer;
    function    LastIndex: Integer;
    function    NextIndex(const CurIndex: Integer): Integer;
    function    PrevIndex(const CurIndex: Integer): Integer;
    procedure   Clear; override;
    function    GetItem(const Index: Integer): Pointer;
    procedure   SetItem(const Index: Integer; const Data: Pointer);
    procedure   ForEachItemDo(const EachItemProc: TEachItemProc);
    function    IndexExists(const Index: Integer): boolean;
    procedure   Pack;
    function    IndexOf(const Data: Pointer): Integer;
    property    Items[const Index: Integer]: Pointer read GetItem write SetItem; default;

    property    Locked: boolean read GetLocked write SetLocked default false;
  end;


implementation

constructor TIndexList.Create;
begin
  inherited Create;
  FListLock:= nil;
end;

destructor TIndexList.Destroy;
begin
  SetLocked(false);
  inherited Destroy;
end;

function TIndexList.GetLocked: boolean;
begin
  Result:= (FListLock <> nil);
end;

procedure TIndexList.SetLocked(const Value: boolean);
begin
  if Value <> GetLocked then
    if Value then
      begin
        New(FListLock);
        InitializeCriticalSection(FListLock^);
      end
    else
      begin
        DeleteCriticalSection(FListLock^);
        Dispose(FListLock);
        FListLock:= nil;
      end;
end;

procedure TIndexList.EnterCriticalSection;
begin
  if GetLocked then Windows.EnterCriticalSection(FListLock^);
end;

procedure TIndexList.LeaveCriticalSection;
begin
  if GetLocked then Windows.LeaveCriticalSection(FListLock^);
end;

function TIndexList.FindNativeIndex(const Index: Integer): Integer;
var
  i: Integer;
begin
  Result:= -1;
  i:= Count-1;
  while i >= 0 do
    if PIndexRec(inherited Items[i])^.Index = Index then
      begin
        Result:= i;
        i:= -1;
      end
    else
      Dec(i);
end;

function TIndexList.Get_FirstIndex: Integer;
begin
  if Count > 0 then
    Result:= PIndexRec(inherited First)^.Index
  else
    Result:= -1;
end;

function TIndexList.Get_LastIndex: Integer;
begin
  if Count > 0 then
    Result:= PIndexRec(inherited Last)^.Index
  else
    Result:= -1;
end;

function TIndexList.Get_NextIndex(const CurIndex: Integer): Integer;
var
  i, j: Integer;
begin
  i:= FindNativeIndex(CurIndex);
  if (i = -1) and (Get_FirstIndex < CurIndex) and (CurIndex < Get_LastIndex) then
    begin
      j:= 1;
      while j < Count do
        if (PIndexRec(inherited Items[j-1])^.Index < CurIndex) and (CurIndex < PIndexRec(inherited Items[j])^.Index) then
          begin
            i:= j-1;
            j:= Count;
          end
        else
          Inc(j);
    end;

  if (i = -1) or (CurIndex = Get_LastIndex) then
    Result:= -1
  else
    Result:= PIndexRec(inherited Items[i+1])^.Index;
end;

function TIndexList.Get_PrevIndex(const CurIndex: Integer): Integer;
var
  i, j: Integer;
begin
  i:= FindNativeIndex(CurIndex);
  if (i = -1) and (Get_FirstIndex < CurIndex) and (CurIndex < Get_LastIndex) then
    begin
      j:= Count-1;
      while j > 0 do
        if (PIndexRec(inherited Items[j-1])^.Index < CurIndex) and (CurIndex < PIndexRec(inherited Items[j])^.Index) then
          begin
            i:= j;
            j:= 0;
          end
        else
          Dec(j);
    end;

  if (i = -1) or (CurIndex = Get_FirstIndex) then
    Result:= -1
  else
    Result:= PIndexRec(inherited Items[i-1])^.Index;
end;

function TIndexList.Add(const Data: Pointer): Integer;
var
  IndexRec: PIndexRec;
begin
  EnterCriticalSection;
  try
    Result:= Get_LastIndex+1;
    New(IndexRec);
    IndexRec^.Index:= Result;
    IndexRec^.Data:= Data;
    inherited Add(IndexRec);
  finally
    LeaveCriticalSection;
  end;
end;

procedure TIndexList.Delete(const Index: Integer);
var
  i: Integer;
begin
  EnterCriticalSection;
  try
    i:= FindNativeIndex(Index);
    if i <> -1 then
      begin
        Dispose(PIndexRec(inherited Items[i]));
        inherited Delete(i);
      end
    else
      raise EIndexListError.Create('IndexList index out of bounds ('+inttostr(Index)+')');
  finally
    LeaveCriticalSection;
  end;
end;

procedure TIndexList.Insert(const Index: Integer; const Data: Pointer);
var
  IndexRec: PIndexRec;
  ExistIndex, i, NativeIndex: Integer;
begin
  EnterCriticalSection;
  try
    if Index >= 0 then
      begin
        New(IndexRec);
        IndexRec^.Index:= Index;
        IndexRec^.Data:= Data;
        ExistIndex:= FindNativeIndex(Index);
        if ExistIndex = -1 then
          if Index > Get_LastIndex then
            inherited Add(IndexRec)
          else
            if Index < Get_FirstIndex then
              inherited Insert(0, IndexRec)
            else
              begin
                i:= Count-1;
                while i >= 1 do
                  if (PIndexRec(inherited Items[i-1])^.Index < Index) and (Index < PIndexRec(inherited Items[i])^.Index) then
                    begin
                      inherited Insert(i, IndexRec);
                      i:= 0;
                    end
                  else
                    Dec(i);
              end
        else
          begin
            i:= ExistIndex;
            while i < Count do
              begin
                NativeIndex:= i;
                if FindNativeIndex(PIndexRec(inherited Items[NativeIndex])^.Index+1) = -1 then
                  i:= Count
                else
                  Inc(i);
                Inc(PIndexRec(inherited Items[NativeIndex])^.Index);
              end;
            inherited Insert(ExistIndex, IndexRec);
          end;
      end
    else
      raise EIndexListError.Create('IndexList index out of bounds (< 0)');
  finally
    LeaveCriticalSection;
  end;
end;

procedure TIndexList.Clear;
var
  i: Integer;
begin
  EnterCriticalSection;
  try
    for i:= 0 to inherited Count-1 do Dispose(PIndexRec(inherited Items[i]));
    inherited Clear;
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.GetItem(const Index: Integer): Pointer;
var
  i: Integer;
begin
  EnterCriticalSection;
  try
    i:= FindNativeIndex(Index);
    if i <> -1 then
      Result:= PIndexRec(inherited Items[i])^.Data
    else
      begin
        Result:= nil;
        raise EIndexListError.Create('IndexList index out of bounds ('+inttostr(Index)+')');
      end;  
  finally
    LeaveCriticalSection;
  end;
end;

procedure TIndexList.SetItem(const Index: Integer; const Data: Pointer);
var
  i: Integer;
begin
  EnterCriticalSection;
  try
    i:= FindNativeIndex(Index);
    if i <> -1 then
      PIndexRec(inherited Items[i])^.Data:= Data
    else
      raise EIndexListError.Create('IndexList index out of bounds ('+inttostr(Index)+')');
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.IndexExists(const Index: Integer): boolean;
begin
  EnterCriticalSection;
  try
    Result:= (FindNativeIndex(Index) <> -1);
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.LastIndex: Integer;
begin
  EnterCriticalSection;
  try
    Result:= Get_LastIndex;
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.FirstIndex: Integer;
begin
  EnterCriticalSection;
  try
    Result:= Get_FirstIndex;
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.NextIndex(const CurIndex: Integer): Integer;
begin
  EnterCriticalSection;
  try
    Result:= Get_NextIndex(CurIndex);
  finally
    LeaveCriticalSection;
  end;
end;

function TIndexList.PrevIndex(const CurIndex: Integer): Integer;
begin
  EnterCriticalSection;
  try
    Result:= Get_PrevIndex(CurIndex);
  finally
    LeaveCriticalSection;
  end;
end;

procedure TIndexList.ForEachItemDo(const EachItemProc: TEachItemProc);
var
  i: Integer;
begin
  if assigned(EachItemProc) then
    try
      EnterCriticalSection;
      i:= 0;
      while i < Count do
        begin
          try
            EachItemProc(Self, PIndexRec(inherited Items[i])^.Data, PIndexRec(inherited Items[i])^.Index);
          except
            { nothing }
          end;
          Inc(i);
        end;
    finally
      LeaveCriticalSection;
    end;
end;

procedure TIndexList.Pack;
var
  i: Integer;
begin
  EnterCriticalSection;
  try
    for i:= 0 to inherited Count-1 do
      if PIndexRec(inherited Items[i])^.Data = nil then
        begin
          Dispose(PIndexRec(inherited Items[i]));
          inherited Items[i]:= nil;
        end;
    inherited Pack;
  finally
    LeaveCriticalSection;
  end;    
end;

function TIndexList.IndexOf(const Data: Pointer): Integer;
var
  i: Integer;
begin
  Result:= -1;
  EnterCriticalSection;
  try
    i:= 0;
    while i < inherited Count do
      if PIndexRec(inherited Items[i])^.Data = Data then
        begin
          Result:= PIndexRec(inherited Items[i])^.Index;
          i:= inherited Count;
        end
      else
        Inc(i);
  finally
    LeaveCriticalSection;
  end;
end;

end.



