unit WndUtil;

{$mode objfpc}{$H+}

interface

uses
  Windows, {$IFDEF MINIMAL}MiniClasses{$ELSE}Classes{$ENDIF}, SysUtils;
  
type
  TInstance = record
    Handle: THandle;
    DefWndProc: WndProc;
    WndMethod: TWndMethod;
  end;
  PInstance = ^TInstance;

function  AllocateHwnd(const Method: TWndMethod): THandle;
procedure DeallocateHwnd(const hWnd: THandle);
procedure AssignMethod(const hWnd: THandle; const Method: TWndMethod);
procedure UnassignMethod(const hWnd: THandle);
procedure PurgeDestroyedHwnds;
function  CallDefWindowProc(const hWnd: THandle; var Msg: TMessage): Boolean;

implementation

var
  WndList: TList;
  CS: TCRITICALSECTION;

function FindWindowIndex(const hWnd: THandle): Integer; forward;
function GlobalWndProc(hWnd: THandle; Msg: Cardinal; wParam, lParam: Integer): Integer; stdcall; forward;

procedure EnterCS; inline;
begin
  if IsMultiThread then EnterCriticalSection(CS);
end;

procedure LeaveCS; inline;
begin
  if IsMultiThread then LeaveCriticalSection(CS);
end;

function AllocateHwnd(const Method: TWndMethod): THandle;
var
  UtilWindowClass: TWndClass;
  TempClass: TWndClass;
  ClassRegistered: Boolean;
begin
  UtilWindowClass.style:= 0;
  UtilWindowClass.lpfnWndProc:= @DefWindowProc;
  UtilWindowClass.cbClsExtra:= 0;
  UtilWindowClass.cbWndExtra:= 0;
  UtilWindowClass.hInstance:= 0;
  UtilWindowClass.hIcon:= 0;
  UtilWindowClass.hCursor:= 0;
  UtilWindowClass.hbrBackground:= 0;
  UtilWindowClass.lpszMenuName:= nil;
  UtilWindowClass.lpszClassName:= 'WUtilWindow';
  UtilWindowClass.hInstance:= HInstance;
  ClassRegistered:= GetClassInfo(HInstance, UtilWindowClass.lpszClassName, TempClass);
  if not ClassRegistered or (TempClass.lpfnWndProc <> @DefWindowProc) then
    begin
      if ClassRegistered then Windows.UnregisterClass(UtilWindowClass.lpszClassName, HInstance);
      Windows.RegisterClass(UtilWindowClass);
    end;
  Result:= CreateWindowEx(WS_EX_TOOLWINDOW, UtilWindowClass.lpszClassName,
                          '', WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance, nil);
  AssignMethod(Result, Method);
end;

procedure DeallocateHwnd(const hWnd: THandle);
begin
  UnassignMethod(hWnd);
  DestroyWindow(hWnd);
end;

procedure AssignMethod(const hWnd: THandle; const Method: TWndMethod);
var
  Instance: PInstance;
  Idx: Integer;
begin
  if (hWnd > 0) and Assigned(Method) then
    begin
      EnterCS;
      Idx:= FindWindowIndex(hWnd);
      if Idx > -1 then
        PInstance(WndList[Idx])^.WndMethod:= Method
      else
        begin
          New(Instance);
          Instance^.Handle:= hWnd;
          Instance^.DefWndProc:= WndProc(GetWindowLong(hWnd, GWL_WNDPROC));
          Instance^.WndMethod:= Method;
          WndList.Add(Instance);
          SetWindowLong(hWnd, GWL_WNDPROC, Integer(@GlobalWndProc));
        end;
      LeaveCS;
    end;
end;
    
procedure UnassignMethod(const hWnd: THandle);
var
  Idx: Integer;
begin
  if hWnd > 0 then
    begin
      EnterCS;
      Idx:= FindWindowIndex(hWnd);
      if Idx > -1 then
        begin
          SetWindowLong(hWnd, GWL_WNDPROC, Integer(PInstance(WndList[Idx])^.DefWndProc));
          Dispose(PInstance(WndList[Idx]));
          WndList.Delete(Idx);
        end;
      LeaveCS;
    end;
end;

procedure PurgeDestroyedHwnds;
var
  i: Integer;
begin
  EnterCS;
  i:= 0;
  while i < WndList.Count do
    if IsWindow(PInstance(WndList[i])^.Handle) then
      Inc(i)
    else
      begin
        Dispose(PInstance(WndList[i]));
        WndList.Delete(i);
      end;
  LeaveCS;
end;

function CallDefWindowProc(const hWnd: THandle; var Msg: TMessage): Boolean;
var
  Idx: Integer;
  Proc: WndProc;
  Instance: PInstance;
begin
  EnterCS;
  Idx:= FindWindowIndex(hWnd);
  if Idx > -1 then
    Instance:= WndList[Idx]
  else
    Instance:= nil;
  LeaveCS;

  if Assigned(Instance) then
    begin
      Msg.Result:= CallWindowProc(Instance^.DefWndProc, hWnd, Msg.Msg, Msg.wParam, Msg.lParam);
      Result:= true;
    end
  else
    Result:= false;
end;

function FindWindowIndex(const hWnd: THandle): Integer;
var
  i: Integer;
begin
  Result:= -1;
  i:= WndList.Count-1;
  while i >= 0 do
    if PInstance(WndList[i])^.Handle = hWnd then
      begin
        Result:= i;
        i:= -1;
      end
    else
      Dec(i);
end;

function GlobalWndProc(hWnd: THandle; Msg: Cardinal; wParam, lParam: Integer): Integer; stdcall;
var
  Idx: Integer;
  Instance: PInstance;
  Message: TMessage;
begin
  EnterCS;
  Idx:= FindWindowIndex(hWnd);
  if Idx >= 0 then
    Instance:= WndList[Idx]
  else
    Instance:= nil;
  LeaveCS;
  
  if Assigned(Instance) then
    begin
      Message.Msg:= Msg;
      Message.wParam:= wParam;
      Message.lParam:= lParam;
      Message.Result:= 0;
      try
        Instance^.WndMethod(Message);
      except
        // nothing
      end;
      Result:= Message.Result;
    end
  else
    Result:= DefWindowProc(hWnd, Msg, wParam, lParam);
end;

var
  i: Integer;
  
initialization
  InitializeCriticalSection(CS);
  WndList:= TList.Create;
  
finalization
  EnterCS;
  for i:= 0 to WndList.Count-1 do
    begin
      SetWindowLong(PInstance(WndList[i])^.Handle, GWL_WNDPROC, Integer(PInstance(WndList[i])^.DefWndProc));
      Dispose(PInstance(WndList[i]));
    end;
  WndList.Free;
  WndList:= nil;
  LeaveCS;
  DeleteCriticalSection(CS);
  
end.

